#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2019 Intel Corporation

from scapy.all import *
import unittest
import pkttest

#{ipv4{ipv4}} test
SRC_ADDR_IPV4_1 = "192.168.1.1"
DST_ADDR_IPV4_1 = "192.168.2.1"

#{ipv6{ipv6}} test
SRC_ADDR_IPV6_1 = "1111:0000:0000:0000:0000:0000:0000:0001"
DST_ADDR_IPV6_1 = "2222:0000:0000:0000:0000:0000:0000:0001"

#{ipv4{ipv6}} test
SRC_ADDR_IPV4_2 = "192.168.11.1"
DST_ADDR_IPV4_2 = "192.168.12.1"
SRC_ADDR_IPV6_2 = "1111:0000:0000:0000:0000:0000:0001:0001"
DST_ADDR_IPV6_2 = "2222:0000:0000:0000:0000:0000:0001:0001"

#{ipv6{ipv4}} test
SRC_ADDR_IPV4_3 = "192.168.21.1"
DST_ADDR_IPV4_3 = "192.168.22.1"
SRC_ADDR_IPV6_3 = "1111:0000:0000:0000:0000:0001:0001:0001"
DST_ADDR_IPV6_3 = "2222:0000:0000:0000:0000:0001:0001:0001"

def config():
    return """
#outter-ipv4 inner-ipv4 tunnel mode test
sp ipv4 out esp protect 5 pri 1 \\
src {0}/32 \\
dst {1}/32 \\
sport 0:65535 dport 0:65535

sp ipv4 in esp protect 6 pri 1 \\
src {1}/32 \\
dst {0}/32 \\
sport 0:65535 dport 0:65535

sa out 5 cipher_algo null auth_algo null mode ipv4-tunnel \\
src {0} dst {1}
sa in 6 cipher_algo null auth_algo null mode ipv4-tunnel \\
src {1} dst {0}

rt ipv4 dst {0}/32 port 1
rt ipv4 dst {1}/32 port 0

#outter-ipv6 inner-ipv6 tunnel mode test
sp ipv6 out esp protect 7 pri 1 \\
src {2}/128 \\
dst {3}/128 \\
sport 0:65535 dport 0:65535

sp ipv6 in esp protect 8 pri 1 \\
src {3}/128 \\
dst {2}/128 \\
sport 0:65535 dport 0:65535

sa out 7 cipher_algo null auth_algo null mode ipv6-tunnel \\
src {2} dst {3}
sa in 8 cipher_algo null auth_algo null mode ipv6-tunnel \\
src {3} dst {2}

rt ipv6 dst {2}/128 port 1
rt ipv6 dst {3}/128 port 0

#outter-ipv4 inner-ipv6 tunnel mode test
sp ipv6 out esp protect 9 pri 1 \\
src {4}/128 \\
dst {5}/128 \\
sport 0:65535 dport 0:65535

sp ipv6 in esp protect 10 pri 1 \\
src {5}/128 \\
dst {4}/128 \\
sport 0:65535 dport 0:65535

sa out 9 cipher_algo null auth_algo null mode ipv4-tunnel \\
src {6} dst {7}
sa in 10 cipher_algo null auth_algo null mode ipv4-tunnel \\
src {7} dst {6}

rt ipv6 dst {4}/128 port 1
rt ipv4 dst {7}/32 port 0

#outter-ipv6 inner-ipv4 tunnel mode test
sp ipv4 out esp protect 11 pri 1 \\
src {8}/32 \\
dst {9}/32 \\
sport 0:65535 dport 0:65535

sp ipv4 in esp protect 12 pri 1 \\
src {9}/32 \\
dst {8}/32 \\
sport 0:65535 dport 0:65535

sa out 11 cipher_algo null auth_algo null mode ipv6-tunnel \\
src {10} dst {11}
sa in 12 cipher_algo null auth_algo null mode ipv6-tunnel \\
src {11} dst {10}

rt ipv4 dst {8}/32 port 1
rt ipv6 dst {11}/128 port 0
""".format(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
           SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
           SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2, SRC_ADDR_IPV4_2, DST_ADDR_IPV4_2,
           SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3, SRC_ADDR_IPV6_3, DST_ADDR_IPV6_3)

ECN_ECT0    = 0x02
ECN_ECT1    = 0x01
ECN_CE      = 0x03
DSCP_1      = 0x04
DSCP_3F     = 0xFC

class TestTunnelHeaderReconstruct(unittest.TestCase):
    def setUp(self):
        self.px = pkttest.PacketXfer()
        th = IP(src=DST_ADDR_IPV4_1, dst=SRC_ADDR_IPV4_1)
        self.sa_ipv4v4 = SecurityAssociation(ESP, spi=6, tunnel_header = th)

        th = IPv6(src=DST_ADDR_IPV6_1, dst=SRC_ADDR_IPV6_1)
        self.sa_ipv6v6 = SecurityAssociation(ESP, spi=8, tunnel_header = th)

        th = IP(src=DST_ADDR_IPV4_2, dst=SRC_ADDR_IPV4_2)
        self.sa_ipv4v6 = SecurityAssociation(ESP, spi=10, tunnel_header = th)

        th = IPv6(src=DST_ADDR_IPV6_3, dst=SRC_ADDR_IPV6_3)
        self.sa_ipv6v4 = SecurityAssociation(ESP, spi=12, tunnel_header = th)

    def gen_pkt_plain_ipv4(self, src, dst, tos):
        pkt = IP(src=src, dst=dst, tos=tos)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")
        return pkt

    def gen_pkt_plain_ipv6(self, src, dst, tc):
        pkt = IPv6(src=src, dst=dst, tc=tc)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")
        return pkt

    def gen_pkt_tun_ipv4v4(self, tos_outter, tos_inner):
        pkt = self.gen_pkt_plain_ipv4(DST_ADDR_IPV4_1, SRC_ADDR_IPV4_1,
                                      tos_inner)
        pkt = self.sa_ipv4v4.encrypt(pkt)
        self.assertEqual(pkt[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(pkt[ESP].spi, 6)
        pkt[IP].tos = tos_outter
        return pkt

    def gen_pkt_tun_ipv6v6(self, tc_outter, tc_inner):
        pkt = self.gen_pkt_plain_ipv6(DST_ADDR_IPV6_1, SRC_ADDR_IPV6_1,
                                      tc_inner)
        pkt = self.sa_ipv6v6.encrypt(pkt)
        self.assertEqual(pkt[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(pkt[ESP].spi, 8)
        pkt[IPv6].tc = tc_outter
        return pkt

    def gen_pkt_tun_ipv4v6(self, tos_outter, tc_inner):
        pkt = self.gen_pkt_plain_ipv6(DST_ADDR_IPV6_2, SRC_ADDR_IPV6_2,
                                      tc_inner)
        pkt = self.sa_ipv4v6.encrypt(pkt)
        self.assertEqual(pkt[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(pkt[ESP].spi, 10)
        pkt[IP].tos = tos_outter
        return pkt

    def gen_pkt_tun_ipv6v4(self, tc_outter, tos_inner):
        pkt = self.gen_pkt_plain_ipv4(DST_ADDR_IPV4_3, SRC_ADDR_IPV4_3,
                                      tos_inner)
        pkt = self.sa_ipv6v4.encrypt(pkt)
        self.assertEqual(pkt[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(pkt[ESP].spi, 12)
        pkt[IPv6].tc = tc_outter
        return pkt

#RFC4301 5.1.2.1 & 5.1.2.2, outbound packets shall be copied ECN field
    def test_outb_ipv4v4_ecn(self):
        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
                                      ECN_ECT1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)
        self.assertEqual(resp[IP].tos, ECN_ECT1)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
                                      ECN_ECT0)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)
        self.assertEqual(resp[IP].tos, ECN_ECT0)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
                                      ECN_CE)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)
        self.assertEqual(resp[IP].tos, ECN_CE)

    def test_outb_ipv6v6_ecn(self):
        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
                                      ECN_ECT1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT1)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
                                      ECN_ECT0)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 7)
        self.assertEqual(resp[IPv6].tc, ECN_ECT0)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
                                      ECN_CE)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 7)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

    def test_outb_ipv4v6_ecn(self):
        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2,
                                      ECN_ECT1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[IP].tos, ECN_ECT1)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2,
                                      ECN_ECT0)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[IP].tos, ECN_ECT0)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2,
                                      ECN_CE)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[IP].tos, ECN_CE)

    def test_outb_ipv6v4_ecn(self):
        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3,
                                      ECN_ECT1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT1)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3,
                                      ECN_ECT0)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT0)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3,
                                      ECN_CE)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

#RFC4301 5.1.2.1 & 5.1.2.2, if outbound packets ECN is CE (0x3), inbound packets
#ECN is overwritten to CE, otherwise no change

#Outter header not CE, Inner header should be no change
    def test_inb_ipv4v4_ecn_inner_no_change(self):
        pkt = self.gen_pkt_tun_ipv4v4(ECN_ECT1, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_ECT0)

        pkt = self.gen_pkt_tun_ipv4v4(ECN_ECT0, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_ECT1)

        pkt = self.gen_pkt_tun_ipv4v4(ECN_ECT1, ECN_CE)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

    def test_inb_ipv6v6_ecn_inner_no_change(self):
        pkt = self.gen_pkt_tun_ipv6v6(ECN_ECT1, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT0)

        pkt = self.gen_pkt_tun_ipv6v6(ECN_ECT0, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT1)

        pkt = self.gen_pkt_tun_ipv6v6(ECN_ECT1, ECN_CE)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

    def test_inb_ipv4v6_ecn_inner_no_change(self):
        pkt = self.gen_pkt_tun_ipv4v6(ECN_ECT1, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT0)

        pkt = self.gen_pkt_tun_ipv4v6(ECN_ECT0, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_ECT1)

        pkt = self.gen_pkt_tun_ipv4v6(ECN_ECT1, ECN_CE)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

    def test_inb_ipv6v4_ecn_inner_no_change(self):
        pkt = self.gen_pkt_tun_ipv6v4(ECN_ECT1, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_ECT0)

        pkt = self.gen_pkt_tun_ipv6v4(ECN_ECT0, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_ECT1)

        pkt = self.gen_pkt_tun_ipv6v4(ECN_ECT1, ECN_CE)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

#Outter header CE, Inner header should be changed to CE
    def test_inb_ipv4v4_ecn_inner_change(self):
        pkt = self.gen_pkt_tun_ipv4v4(ECN_CE, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

        pkt = self.gen_pkt_tun_ipv4v4(ECN_CE, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

    def test_inb_ipv6v6_ecn_inner_change(self):
        pkt = self.gen_pkt_tun_ipv6v6(ECN_CE, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

        pkt = self.gen_pkt_tun_ipv6v6(ECN_CE, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

    def test_inb_ipv4v6_ecn_inner_change(self):
        pkt = self.gen_pkt_tun_ipv4v6(ECN_CE, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

        pkt = self.gen_pkt_tun_ipv4v6(ECN_CE, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, ECN_CE)

    def test_inb_ipv6v4_ecn_inner_change(self):
        pkt = self.gen_pkt_tun_ipv6v4(ECN_CE, ECN_ECT0)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

        pkt = self.gen_pkt_tun_ipv6v4(ECN_CE, ECN_ECT1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, ECN_CE)

#RFC4301 5.1.2.1.5 Outer DS field should be copied from Inner DS field
    def test_outb_ipv4v4_dscp(self):
        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
                                      DSCP_1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)
        self.assertEqual(resp[IP].tos, DSCP_1)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_1, DST_ADDR_IPV4_1,
                                      DSCP_3F)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)
        self.assertEqual(resp[IP].tos, DSCP_3F)

    def test_outb_ipv6v6_dscp(self):
        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
                                      DSCP_1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 7)
        self.assertEqual(resp[IPv6].tc, DSCP_1)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_1, DST_ADDR_IPV6_1,
                                      DSCP_3F)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 7)
        self.assertEqual(resp[IPv6].tc, DSCP_3F)

    def test_outb_ipv4v6_dscp(self):
        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2,
                                      DSCP_1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 9)
        self.assertEqual(resp[IP].tos, DSCP_1)

        pkt = self.gen_pkt_plain_ipv6(SRC_ADDR_IPV6_2, DST_ADDR_IPV6_2,
                                      DSCP_3F)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 9)
        self.assertEqual(resp[IP].tos, DSCP_3F)

    def test_outb_ipv6v4_dscp(self):
        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3,
                                      DSCP_1)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 11)
        self.assertEqual(resp[IPv6].tc, DSCP_1)

        pkt = self.gen_pkt_plain_ipv4(SRC_ADDR_IPV4_3, DST_ADDR_IPV4_3,
                                      DSCP_3F)
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 11)
        self.assertEqual(resp[IPv6].tc, DSCP_3F)

#RFC4301 5.1.2.1.5 Inner DS field should not be affected by Outer DS field
    def test_inb_ipv4v4_dscp(self):
        pkt = self.gen_pkt_tun_ipv4v4(DSCP_3F, DSCP_1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, DSCP_1)

        pkt = self.gen_pkt_tun_ipv4v4(DSCP_1, DSCP_3F)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, DSCP_3F)

    def test_inb_ipv6v6_dscp(self):
        pkt = self.gen_pkt_tun_ipv6v6(DSCP_3F, DSCP_1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, DSCP_1)

        pkt = self.gen_pkt_tun_ipv6v6(DSCP_1, DSCP_3F)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, DSCP_3F)

    def test_inb_ipv4v6_dscp(self):
        pkt = self.gen_pkt_tun_ipv4v6(DSCP_3F, DSCP_1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, DSCP_1)

        pkt = self.gen_pkt_tun_ipv4v6(DSCP_1, DSCP_3F)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(resp[IPv6].tc, DSCP_3F)

    def test_inb_ipv6v4_dscp(self):
        pkt = self.gen_pkt_tun_ipv6v4(DSCP_3F, DSCP_1)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, DSCP_1)

        pkt = self.gen_pkt_tun_ipv6v4(DSCP_1, DSCP_3F)
        resp = self.px.xfer_protected(pkt)
        self.assertEqual(resp[IP].proto, socket.IPPROTO_UDP)
        self.assertEqual(resp[IP].tos, DSCP_3F)

pkttest.pkttest()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2016  Neil Horman <nhorman@tuxdriver.com>
# Copyright(c) 2022  Robin Jarry
# pylint: disable=invalid-name

r"""
Utility to dump PMD_INFO_STRING support from DPDK binaries.

This script prints JSON output to be interpreted by other tools. Here are some
examples with jq:

Get the complete info for a given driver:

  %(prog)s dpdk-testpmd | \
  jq '.[] | select(.name == "cnxk_nix_inl")'

Get only the required kernel modules for a given driver:

  %(prog)s dpdk-testpmd | \
  jq '.[] | select(.name == "net_i40e").kmod'

Get only the required kernel modules for a given device:

  %(prog)s dpdk-testpmd | \
  jq '.[] | select(.pci_ids[]? | .vendor == "15b3" and .device == "1013").kmod'
"""

import argparse
import json
import logging
import os
import re
import string
import subprocess
import sys
from pathlib import Path
from typing import Iterable, Iterator, List, Union

import elftools
from elftools.elf.elffile import ELFError, ELFFile


# ----------------------------------------------------------------------------
def main() -> int:  # pylint: disable=missing-docstring
    try:
        args = parse_args()
        logging.basicConfig(
            stream=sys.stderr,
            format="%(levelname)s: %(message)s",
            level={
                0: logging.ERROR,
                1: logging.WARNING,
            }.get(args.verbose, logging.DEBUG),
        )
        info = parse_pmdinfo(args.elf_files, args.search_plugins)
        print(json.dumps(info, indent=2))
    except BrokenPipeError:
        pass
    except KeyboardInterrupt:
        return 1
    except Exception as e:  # pylint: disable=broad-except
        logging.error("%s", e)
        return 1
    return 0


# ----------------------------------------------------------------------------
def parse_args() -> argparse.Namespace:
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-p",
        "--search-plugins",
        action="store_true",
        help="""
        In addition of ELF_FILEs and their linked dynamic libraries, also scan
        the DPDK plugins path.
        """,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="""
        Display warnings due to linked libraries not found or ELF/JSON parsing
        errors in these libraries. Use twice to show debug messages.
        """,
    )
    parser.add_argument(
        "elf_files",
        metavar="ELF_FILE",
        nargs="+",
        type=existing_file,
        help="""
        DPDK application binary or dynamic library.
        """,
    )
    return parser.parse_args()


# ----------------------------------------------------------------------------
def parse_pmdinfo(paths: Iterable[Path], search_plugins: bool) -> List[dict]:
    """
    Extract DPDK PMD info JSON strings from an ELF file.

    :returns:
        A list of DPDK drivers info dictionaries.
    """
    binaries = set(paths)
    for p in paths:
        binaries.update(get_needed_libs(p))
    if search_plugins:
        # cast to list to avoid errors with update while iterating
        binaries.update(list(get_plugin_libs(binaries)))

    drivers = []

    for b in binaries:
        logging.debug("analyzing %s", b)
        try:
            for s in get_elf_strings(b, ".rodata", "PMD_INFO_STRING="):
                try:
                    info = json.loads(s)
                    scrub_pci_ids(info)
                    drivers.append(info)
                except ValueError as e:
                    # invalid JSON, should never happen
                    logging.warning("%s: %s", b, e)
        except ELFError as e:
            # only happens for discovered plugins that are not ELF
            logging.debug("%s: cannot parse ELF: %s", b, e)

    return drivers


# ----------------------------------------------------------------------------
PCI_FIELDS = ("vendor", "device", "subsystem_vendor", "subsystem_device")


def scrub_pci_ids(info: dict):
    """
    Convert numerical ids to hex strings.
    Strip empty pci_ids lists.
    Strip wildcard 0xFFFF ids.
    """
    pci_ids = []
    for pci_fields in info.pop("pci_ids"):
        pci = {}
        for name, value in zip(PCI_FIELDS, pci_fields):
            if value != 0xFFFF:
                pci[name] = f"{value:04x}"
        if pci:
            pci_ids.append(pci)
    if pci_ids:
        info["pci_ids"] = pci_ids


# ----------------------------------------------------------------------------
def get_plugin_libs(binaries: Iterable[Path]) -> Iterator[Path]:
    """
    Look into the provided binaries for DPDK_PLUGIN_PATH and scan the path
    for files.
    """
    for b in binaries:
        for p in get_elf_strings(b, ".rodata", "DPDK_PLUGIN_PATH="):
            plugin_path = p.strip()
            logging.debug("discovering plugins in %s", plugin_path)
            for root, _, files in os.walk(plugin_path):
                for f in files:
                    yield Path(root) / f
            # no need to search in other binaries.
            return


# ----------------------------------------------------------------------------
def existing_file(value: str) -> Path:
    """
    Argparse type= callback to ensure an argument points to a valid file path.
    """
    path = Path(value)
    if not path.is_file():
        raise argparse.ArgumentTypeError(f"{value}: No such file")
    return path


# ----------------------------------------------------------------------------
PRINTABLE_BYTES = frozenset(string.printable.encode("ascii"))


def find_strings(buf: bytes, prefix: str) -> Iterator[str]:
    """
    Extract strings of printable ASCII characters from a bytes buffer.
    """
    view = memoryview(buf)
    start = None

    for i, b in enumerate(view):
        if start is None and b in PRINTABLE_BYTES:
            # mark beginning of string
            start = i
            continue
        if start is not None:
            if b in PRINTABLE_BYTES:
                # string not finished
                continue
            if b == 0:
                # end of string
                s = view[start:i].tobytes().decode("ascii")
                if s.startswith(prefix):
                    yield s[len(prefix):]
            # There can be byte sequences where a non-printable byte
            # follows a printable one. Ignore that.
            start = None


# ----------------------------------------------------------------------------
def elftools_version():
    """
    Extract pyelftools version as a tuple of integers for easy comparison.
    """
    version = getattr(elftools, "__version__", "")
    match = re.match(r"^(\d+)\.(\d+).*$", str(version))
    if not match:
        # cannot determine version, hope for the best
        return (0, 24)
    return (int(match[1]), int(match[2]))


ELFTOOLS_VERSION = elftools_version()


def from_elftools(s: Union[bytes, str]) -> str:
    """
    Earlier versions of pyelftools (< 0.24) return bytes encoded with "latin-1"
    instead of python strings.
    """
    if isinstance(s, bytes):
        return s.decode("latin-1")
    return s


def to_elftools(s: str) -> Union[bytes, str]:
    """
    Earlier versions of pyelftools (< 0.24) assume that ELF section and tags
    are bytes encoded with "latin-1" instead of python strings.
    """
    if ELFTOOLS_VERSION < (0, 24):
        return s.encode("latin-1")
    return s


# ----------------------------------------------------------------------------
def get_elf_strings(path: Path, section: str, prefix: str) -> Iterator[str]:
    """
    Extract strings from a named ELF section in a file.
    """
    with path.open("rb") as f:
        elf = ELFFile(f)
        sec = elf.get_section_by_name(to_elftools(section))
        if not sec:
            return
        yield from find_strings(sec.data(), prefix)


# ----------------------------------------------------------------------------
LDD_LIB_RE = re.compile(
    r"""
    ^                  # beginning of line
    \t                 # tab
    (\S+)              # lib name
    \s+=>\s+
    (/\S+)             # lib path
    \s+
    \(0x[0-9A-Fa-f]+\) # address
    \s*
    $                  # end of line
    """,
    re.MULTILINE | re.VERBOSE,
)


def get_needed_libs(path: Path) -> Iterator[Path]:
    """
    Extract the dynamic library dependencies from an ELF executable.
    """
    with subprocess.Popen(
        ["ldd", str(path)], stdout=subprocess.PIPE, stderr=subprocess.PIPE
    ) as proc:
        out, err = proc.communicate()
        if proc.returncode != 0:
            err = err.decode("utf-8").splitlines()[-1].strip()
            raise Exception(f"cannot read ELF file: {err}")
        for match in LDD_LIB_RE.finditer(out.decode("utf-8")):
            libname, libpath = match.groups()
            if libname.startswith("librte_"):
                libpath = Path(libpath)
                if libpath.is_file():
                    yield libpath.resolve()


# ----------------------------------------------------------------------------
if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2014 6WIND S.A.
# Copyright (c) 2023 Robin Jarry

"""
Craft IP{v6}/{TCP/UDP} traffic flows that will evenly spread over a given
number of RX queues according to the RSS algorithm.
"""

import argparse
import binascii
import ctypes
import ipaddress
import json
import struct
import typing


Address = typing.Union[ipaddress.IPv4Address, ipaddress.IPv6Address]
Network = typing.Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
PortList = typing.Iterable[int]


class Packet:
    def __init__(self, ip_src: Address, ip_dst: Address, l4_sport: int, l4_dport: int):
        self.ip_src = ip_src
        self.ip_dst = ip_dst
        self.l4_sport = l4_sport
        self.l4_dport = l4_dport

    def reverse(self):
        return Packet(
            ip_src=self.ip_dst,
            l4_sport=self.l4_dport,
            ip_dst=self.ip_src,
            l4_dport=self.l4_sport,
        )

    def hash_data(self, use_l4_port: bool = False) -> bytes:
        data = self.ip_src.packed + self.ip_dst.packed
        if use_l4_port:
            data += struct.pack(">H", self.l4_sport)
            data += struct.pack(">H", self.l4_dport)
        return data


class TrafficTemplate:
    def __init__(
        self,
        ip_src: Network,
        ip_dst: Network,
        l4_sport_range: PortList,
        l4_dport_range: PortList,
    ):
        self.ip_src = ip_src
        self.ip_dst = ip_dst
        self.l4_sport_range = l4_sport_range
        self.l4_dport_range = l4_dport_range

    def __iter__(self) -> typing.Iterator[Packet]:
        for ip_src in self.ip_src.hosts():
            for ip_dst in self.ip_dst.hosts():
                if ip_src == ip_dst:
                    continue
                for sport in self.l4_sport_range:
                    for dport in self.l4_dport_range:
                        yield Packet(ip_src, ip_dst, sport, dport)


class RSSAlgo:
    def __init__(
        self,
        queues_count: int,
        key: bytes,
        reta_size: int,
        use_l4_port: bool,
    ):
        self.queues_count = queues_count
        self.reta = tuple(i % queues_count for i in range(reta_size))
        self.key = key
        self.use_l4_port = use_l4_port

    def toeplitz_hash(self, data: bytes) -> int:
        # see rte_softrss_* in lib/hash/rte_thash.h
        hash_value = ctypes.c_uint32(0)

        for i, byte in enumerate(data):
            for j in range(8):
                bit = (byte >> (7 - j)) & 0x01

                if bit == 1:
                    keyword = ctypes.c_uint32(0)
                    keyword.value |= self.key[i] << 24
                    keyword.value |= self.key[i + 1] << 16
                    keyword.value |= self.key[i + 2] << 8
                    keyword.value |= self.key[i + 3]

                    if j > 0:
                        keyword.value <<= j
                        keyword.value |= self.key[i + 4] >> (8 - j)

                    hash_value.value ^= keyword.value

        return hash_value.value

    def get_queue_index(self, packet: Packet) -> int:
        bytes_to_hash = packet.hash_data(self.use_l4_port)

        # get the 32bit hash of the packet
        hash_value = self.toeplitz_hash(bytes_to_hash)

        # determine the offset in the redirection table
        offset = hash_value & (len(self.reta) - 1)

        return self.reta[offset]


def balanced_traffic(
    algo: RSSAlgo,
    traffic_template: TrafficTemplate,
    check_reverse_traffic: bool = False,
    all_flows: bool = False,
) -> typing.Iterator[typing.Tuple[int, int, Packet]]:
    queues = set()
    if check_reverse_traffic:
        queues_reverse = set()

    for pkt in traffic_template:
        q = algo.get_queue_index(pkt)

        # check if q is already filled
        if not all_flows and q in queues:
            continue

        qr = algo.get_queue_index(pkt.reverse())

        if check_reverse_traffic:
            # check if q is already filled
            if not all_flows and qr in queues_reverse:
                continue
            # mark this queue as matched
            queues_reverse.add(qr)

        # mark this queue as filled
        queues.add(q)

        yield q, qr, pkt

        # stop when all queues have been filled
        if not all_flows and len(queues) == algo.queues_count:
            break


NO_PORT = (0,)


class DriverInfo:
    def __init__(self, key: bytes = None, reta_size: int = None):
        self.__key = key
        self.__reta_size = reta_size

    def rss_key(self) -> bytes:
        return self.__key

    def reta_size(self, num_queues: int) -> int:
        return self.__reta_size


class MlxDriverInfo(DriverInfo):
    def rss_key(self) -> bytes:
        return bytes(
            (
                # fmt: off
                # rss_hash_default_key, see drivers/net/mlx5/mlx5_rxq.c
                0x2c, 0xc6, 0x81, 0xd1, 0x5b, 0xdb, 0xf4, 0xf7,
                0xfc, 0xa2, 0x83, 0x19, 0xdb, 0x1a, 0x3e, 0x94,
                0x6b, 0x9e, 0x38, 0xd9, 0x2c, 0x9c, 0x03, 0xd1,
                0xad, 0x99, 0x44, 0xa7, 0xd9, 0x56, 0x3d, 0x59,
                0x06, 0x3c, 0x25, 0xf3, 0xfc, 0x1f, 0xdc, 0x2a,
                # fmt: on
            )
        )

    def reta_size(self, num_queues: int) -> int:
        if num_queues & (num_queues - 1) == 0:
            # If the requested number of RX queues is power of two,
            # use a table of this size.
            return num_queues
        # otherwise, use the maximum table size
        return 512


DEFAULT_DRIVERS = {
    "cnxk": DriverInfo(
        key=bytes(
            (
                # fmt: off
                # roc_nix_rss_key_default_fill, see drivers/common/cnxk/roc_nix_rss.c
                # Marvell cnxk NICs take 48 bytes keys
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                0xfe, 0xed, 0x0b, 0xad, 0xfe, 0xed, 0x0b, 0xad,
                # fmt: on
            )
        ),
        reta_size=64,
    ),
    "intel": DriverInfo(
        key=bytes(
            (
                # fmt: off
                # rss_intel_key, see drivers/net/ixgbe/ixgbe_rxtx.c
                0x6d, 0x5a, 0x56, 0xda, 0x25, 0x5b, 0x0e, 0xc2,
                0x41, 0x67, 0x25, 0x3d, 0x43, 0xa3, 0x8f, 0xb0,
                0xd0, 0xca, 0x2b, 0xcb, 0xae, 0x7b, 0x30, 0xb4,
                0x77, 0xcb, 0x2d, 0xa3, 0x80, 0x30, 0xf2, 0x0c,
                0x6a, 0x42, 0xb7, 0x3b, 0xbe, 0xac, 0x01, 0xfa,
                # fmt: on
            )
        ),
        reta_size=128,
    ),
    "i40e": DriverInfo(
        key=bytes(
            (
                # fmt: off
                # rss_key_default, see drivers/net/i40e/i40e_ethdev.c
                # i40e is the only driver that takes 52 bytes keys
                0x44, 0x39, 0x79, 0x6b, 0xb5, 0x4c, 0x50, 0x23,
                0xb6, 0x75, 0xea, 0x5b, 0x12, 0x4f, 0x9f, 0x30,
                0xb8, 0xa2, 0xc0, 0x3d, 0xdf, 0xdc, 0x4d, 0x02,
                0xa0, 0x8c, 0x9b, 0x33, 0x4a, 0xf6, 0x4a, 0x4c,
                0x05, 0xc6, 0xfa, 0x34, 0x39, 0x58, 0xd8, 0x55,
                0x7d, 0x99, 0x58, 0x3a, 0xe1, 0x38, 0xc9, 0x2e,
                0x81, 0x15, 0x03, 0x66,
                # fmt: on
            )
        ),
        reta_size=512,
    ),
    "mlx": MlxDriverInfo(),
}


def port_range(value):
    try:
        if "-" in value:
            start, stop = value.split("-")
            res = tuple(range(int(start), int(stop)))
        else:
            res = (int(value),)
        return res or NO_PORT
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e)) from e


def positive_int(value):
    try:
        i = int(value)
        if i <= 0:
            raise argparse.ArgumentTypeError("must be strictly positive")
        return i
    except ValueError as e:
        raise argparse.ArgumentTypeError(str(e)) from e


def power_of_two(value):
    i = positive_int(value)
    if i & (i - 1) != 0:
        raise argparse.ArgumentTypeError("must be a power of two")
    return i


def parse_args():
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        "rx_queues",
        metavar="RX_QUEUES",
        type=positive_int,
        help="""
        The number of RX queues to fill.
        """,
    )
    parser.add_argument(
        "ip_src",
        metavar="SRC",
        type=ipaddress.ip_network,
        help="""
        The source IP network/address.
        """,
    )
    parser.add_argument(
        "ip_dst",
        metavar="DST",
        type=ipaddress.ip_network,
        help="""
        The destination IP network/address.
        """,
    )
    parser.add_argument(
        "-s",
        "--sport-range",
        type=port_range,
        default=NO_PORT,
        help="""
        The layer 4 (TCP/UDP) source port range.
        Can be a single fixed value or a range <start>-<end>.
        """,
    )
    parser.add_argument(
        "-d",
        "--dport-range",
        type=port_range,
        default=NO_PORT,
        help="""
        The layer 4 (TCP/UDP) destination port range.
        Can be a single fixed value or a range <start>-<end>.
        """,
    )
    parser.add_argument(
        "-r",
        "--check-reverse-traffic",
        action="store_true",
        help="""
        The reversed traffic (source <-> dest) should also be evenly balanced
        in the queues.
        """,
    )
    parser.add_argument(
        "-k",
        "--rss-key",
        default="intel",
        help=f"""
        The random key used to compute the RSS hash. This option
        supports either a well-known name or the hex value of the key
        (well-known names: {', '.join(DEFAULT_DRIVERS)}, default: intel).
        """,
    )
    parser.add_argument(
        "-t",
        "--reta-size",
        type=power_of_two,
        help="""
        Size of the redirection table or "RETA" (default: depends on driver if
        using a well-known driver name, otherwise 128).
        """,
    )
    parser.add_argument(
        "-a",
        "--all-flows",
        action="store_true",
        help="""
        Output ALL flows that can be created based on source and destination
        address/port ranges along their matched queue number. ATTENTION: this
        option can produce very long outputs depending on the address and port
        range sizes.
        """,
    )
    parser.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="""
        Output in parseable JSON format.
        """,
    )
    parser.add_argument(
        "-i",
        "--info",
        action="store_true",
        help="""
        Print RETA size and RSS key above the results. Not available with --json.
        """,
    )

    args = parser.parse_args()

    if args.ip_src.version != args.ip_dst.version:
        parser.error(
            f"{args.ip_src} and {args.ip_dst} don't have the same protocol version"
        )

    if args.json and args.info:
        parser.error("--json and --info are mutually exclusive")

    if args.rss_key in DEFAULT_DRIVERS:
        driver_info = DEFAULT_DRIVERS[args.rss_key]
    else:
        try:
            key = binascii.unhexlify(args.rss_key)
        except (TypeError, ValueError) as e:
            parser.error(f"RSS_KEY: {e}")
        driver_info = DriverInfo(key=key, reta_size=128)

    if args.reta_size is None:
        args.reta_size = driver_info.reta_size(args.rx_queues)

    if args.reta_size < args.rx_queues:
        parser.error("RETA_SIZE must be greater than or equal to RX_QUEUES")

    args.rss_key = driver_info.rss_key()

    return args


def main():
    args = parse_args()
    use_l4_port = args.sport_range != NO_PORT or args.dport_range != NO_PORT

    algo = RSSAlgo(
        queues_count=args.rx_queues,
        key=args.rss_key,
        reta_size=args.reta_size,
        use_l4_port=use_l4_port,
    )
    template = TrafficTemplate(
        args.ip_src,
        args.ip_dst,
        args.sport_range,
        args.dport_range,
    )

    results = balanced_traffic(
        algo, template, args.check_reverse_traffic, args.all_flows
    )

    if args.json:
        flows = []
        for q, qr, pkt in results:
            flows.append(
                {
                    "queue": q,
                    "queue_reverse": qr,
                    "src_ip": str(pkt.ip_src),
                    "dst_ip": str(pkt.ip_dst),
                    "src_port": pkt.l4_sport,
                    "dst_port": pkt.l4_dport,
                }
            )
        print(json.dumps(flows, indent=2))
        return

    if use_l4_port:
        header = ["SRC_IP", "SPORT", "DST_IP", "DPORT", "QUEUE"]
    else:
        header = ["SRC_IP", "DST_IP", "QUEUE"]
    if args.check_reverse_traffic:
        header.append("QUEUE_REVERSE")

    rows = [tuple(header)]
    widths = [len(h) for h in header]

    for q, qr, pkt in results:
        if use_l4_port:
            row = [pkt.ip_src, pkt.l4_sport, pkt.ip_dst, pkt.l4_dport, q]
        else:
            row = [pkt.ip_src, pkt.ip_dst, q]
        if args.check_reverse_traffic:
            row.append(qr)
        cells = []
        for i, r in enumerate(row):
            r = str(r)
            if len(r) > widths[i]:
                widths[i] = len(r)
            cells.append(r)
        rows.append(tuple(cells))

    if args.info:
        print(f"RSS key:     {binascii.hexlify(args.rss_key).decode()}")
        print(f"RETA size:   {args.reta_size}")
        print()

    fmt = [f"%-{w}s" for w in widths]
    fmt[-1] = "%s"  # avoid trailing whitespace
    fmt = "    ".join(fmt)
    for row in rows:
        print(fmt % row)


if __name__ == "__main__":
    main()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Robin Jarry

CPU_TOTAL = "total_cycles"
CPU_BUSY = "busy_cycles"


def info() -> "dict[Name, tuple[Description, Type]]":
    return {
        CPU_TOTAL: ("Total number of CPU cycles.", "counter"),
        CPU_BUSY: ("Number of busy CPU cycles.", "counter"),
    }


def metrics(sock: "TelemetrySocket") -> "list[tuple[Name, Value, Labels]]":
    out = []
    for lcore_id in sock.cmd("/eal/lcore/list"):
        lcore = sock.cmd("/eal/lcore/info", lcore_id)
        cpu = ",".join(str(c) for c in lcore.get("cpuset", []))
        total = lcore.get("total_cycles")
        busy = lcore.get("busy_cycles", 0)
        if not (cpu and total):
            continue
        labels = {"cpu": cpu, "numa": lcore.get("socket", 0)}
        out += [
            (CPU_TOTAL, total, labels),
            (CPU_BUSY, busy, labels),
        ]
    return out



#! /usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2020 Microsoft Corporation
"""Script to query and setup huge pages for DPDK applications."""

import argparse
import glob
import os
import re
import sys
from math import log2

# Standard binary prefix
BINARY_PREFIX = "KMG"

# systemd mount point for huge pages
HUGE_MOUNT = "/dev/hugepages"


def fmt_memsize(kb):
    '''Format memory size in kB into conventional format'''
    logk = int(log2(kb) / 10)
    suffix = BINARY_PREFIX[logk]
    unit = 2**(logk * 10)
    return '{}{}b'.format(int(kb / unit), suffix)


def get_memsize(arg):
    '''Convert memory size with suffix to kB'''
    match = re.match(r'(\d+)([' + BINARY_PREFIX + r']?)$', arg.upper())
    if match is None:
        sys.exit('{} is not a valid size'.format(arg))
    num = float(match.group(1))
    suffix = match.group(2)
    if suffix == "":
        return int(num / 1024)
    idx = BINARY_PREFIX.find(suffix)
    return int(num * (2**(idx * 10)))


def is_numa():
    '''Test if NUMA is necessary on this system'''
    return os.path.exists('/sys/devices/system/node')


def get_valid_page_sizes(path):
    '''Extract valid hugepage sizes'''
    dir = os.path.dirname(path)
    pg_sizes = (d.split("-")[1] for d in os.listdir(dir))
    return " ".join(pg_sizes)


def get_hugepages(path):
    '''Read number of reserved pages'''
    with open(path + '/nr_hugepages') as nr_hugepages:
        return int(nr_hugepages.read())
    return 0


def set_hugepages(path, reqpages):
    '''Write the number of reserved huge pages'''
    filename = path + '/nr_hugepages'
    try:
        with open(filename, 'w') as nr_hugepages:
            nr_hugepages.write('{}\n'.format(reqpages))
    except PermissionError:
        sys.exit('Permission denied: need to be root!')
    except FileNotFoundError:
        sys.exit("Invalid page size. Valid page sizes: {}".format(
                 get_valid_page_sizes(path)))
    gotpages = get_hugepages(path)
    if gotpages != reqpages:
        sys.exit('Unable to set pages ({} instead of {} in {}).'.format(
                 gotpages, reqpages, filename))


def show_numa_pages():
    '''Show huge page reservations on Numa system'''
    print('Node Pages Size Total')
    for numa_path in glob.glob('/sys/devices/system/node/node*'):
        node = numa_path[29:]  # slice after /sys/devices/system/node/node
        path = numa_path + '/hugepages'
        if not os.path.exists(path):
            continue
        for hdir in os.listdir(path):
            pages = get_hugepages(path + '/' + hdir)
            if pages > 0:
                kb = int(hdir[10:-2])  # slice out of hugepages-NNNkB
                print('{:<4} {:<5} {:<6} {}'.format(node, pages,
                                                    fmt_memsize(kb),
                                                    fmt_memsize(pages * kb)))


def show_non_numa_pages():
    '''Show huge page reservations on non Numa system'''
    print('Pages Size Total')
    path = '/sys/kernel/mm/hugepages'
    for hdir in os.listdir(path):
        pages = get_hugepages(path + '/' + hdir)
        if pages > 0:
            kb = int(hdir[10:-2])
            print('{:<5} {:<6} {}'.format(pages, fmt_memsize(kb),
                                          fmt_memsize(pages * kb)))


def show_pages():
    '''Show existing huge page settings'''
    if is_numa():
        show_numa_pages()
    else:
        show_non_numa_pages()


def clear_pages():
    '''Clear all existing huge page mappings'''
    if is_numa():
        dirs = glob.glob(
            '/sys/devices/system/node/node*/hugepages/hugepages-*')
    else:
        dirs = glob.glob('/sys/kernel/mm/hugepages/hugepages-*')

    for path in dirs:
        set_hugepages(path, 0)


def default_pagesize():
    '''Get default huge page size from /proc/meminfo'''
    with open('/proc/meminfo') as meminfo:
        for line in meminfo:
            if line.startswith('Hugepagesize:'):
                return int(line.split()[1])
    return None


def set_numa_pages(pages, hugepgsz, node=None):
    '''Set huge page reservation on Numa system'''
    if node:
        nodes = ['/sys/devices/system/node/node{}/hugepages'.format(node)]
    else:
        nodes = glob.glob('/sys/devices/system/node/node*/hugepages')

    for node_path in nodes:
        huge_path = '{}/hugepages-{}kB'.format(node_path, hugepgsz)
        set_hugepages(huge_path, pages)


def set_non_numa_pages(pages, hugepgsz):
    '''Set huge page reservation on non Numa system'''
    path = '/sys/kernel/mm/hugepages/hugepages-{}kB'.format(hugepgsz)
    set_hugepages(path, pages)


def reserve_pages(pages, hugepgsz, node=None):
    '''Set the number of huge pages to be reserved'''
    if node or is_numa():
        set_numa_pages(pages, hugepgsz, node=node)
    else:
        set_non_numa_pages(pages, hugepgsz)


def get_mountpoints():
    '''Get list of where hugepage filesystem is mounted'''
    mounted = []
    with open('/proc/mounts') as mounts:
        for line in mounts:
            fields = line.split()
            if fields[2] != 'hugetlbfs':
                continue
            mounted.append(fields[1])
    return mounted


def mount_huge(pagesize, mountpoint, user, group):
    '''Mount the huge TLB file system'''
    if mountpoint in get_mountpoints():
        print(mountpoint, "already mounted")
        return
    cmd = "mount -t hugetlbfs"
    if pagesize:
        cmd += ' -o pagesize={}'.format(pagesize * 1024)
    if user:
        cmd += ' -o uid=' + user
    if group:
        cmd += ' -o gid=' + group
    cmd += ' nodev ' + mountpoint
    os.system(cmd)


def umount_huge(mountpoint):
    '''Unmount the huge TLB file system (if mounted)'''
    if mountpoint in get_mountpoints():
        os.system("umount " + mountpoint)


def show_mount():
    '''Show where huge page filesystem is mounted'''
    mounted = get_mountpoints()
    if mounted:
        print("Hugepages mounted on", *mounted)
    else:
        print("Hugepages not mounted")


def main():
    '''Process the command line arguments and setup huge pages'''
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="Setup huge pages",
        epilog="""
Examples:

To display current huge page settings:
    %(prog)s -s

To a complete setup of with 2 Gigabyte of 1G huge pages:
    %(prog)s -p 1G --setup 2G
""")
    parser.add_argument(
        '--show',
        '-s',
        action='store_true',
        help="print the current huge page configuration")
    parser.add_argument(
        '--clear', '-c', action='store_true', help="clear existing huge pages")
    parser.add_argument(
        '--mount',
        '-m',
        action='store_true',
        help='mount the huge page filesystem')
    parser.add_argument(
        '--unmount',
        '-u',
        action='store_true',
        help='unmount the system huge page directory')
    parser.add_argument(
        '--directory',
        '-d',
        metavar='DIR',
        default=HUGE_MOUNT,
        help='mount point')
    parser.add_argument(
        '--user',
        '-U',
        metavar='UID',
        help='set the mounted directory owner user')
    parser.add_argument(
        '--group',
        '-G',
        metavar='GID',
        help='set the mounted directory owner group')
    parser.add_argument(
        '--node', '-n', help='select numa node to reserve pages on')
    parser.add_argument(
        '--pagesize',
        '-p',
        metavar='SIZE',
        help='choose huge page size to use')
    parser.add_argument(
        '--reserve',
        '-r',
        metavar='SIZE',
        help='reserve huge pages. Size is in bytes with K, M, or G suffix')
    parser.add_argument(
        '--setup',
        metavar='SIZE',
        help='setup huge pages by doing clear, unmount, reserve and mount')
    args = parser.parse_args()

    if args.setup:
        args.clear = True
        args.unmount = True
        args.reserve = args.setup
        args.mount = True

    if not (args.show or args.mount or args.unmount or args.clear or args.reserve):
        parser.error("no action specified")

    if args.pagesize:
        pagesize_kb = get_memsize(args.pagesize)
    else:
        pagesize_kb = default_pagesize()
    if not pagesize_kb:
        sys.exit("Invalid page size: {}kB".format(pagesize_kb))

    if args.clear:
        clear_pages()
    if args.unmount:
        umount_huge(args.directory)

    if args.reserve:
        reserve_kb = get_memsize(args.reserve)
        if reserve_kb % pagesize_kb != 0:
            sys.exit(
                'Huge reservation {}kB is not a multiple of page size {}kB'.
                format(reserve_kb, pagesize_kb))
        reserve_pages(
            int(reserve_kb / pagesize_kb), pagesize_kb, node=args.node)
    if args.mount:
        mount_huge(pagesize_kb, args.directory, args.user, args.group)
    if args.show:
        show_pages()
        print()
        show_mount()


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause

from scapy.all import *
import unittest
import pkttest


SRC_ADDR  = "1111:0000:0000:0000:0000:0000:0000:0001"
DST_ADDR  = "2222:0000:0000:0000:0000:0000:0000:0001"
SRC_NET   = "1111:0000:0000:0000:0000:0000:0000:0000/64"
DST_NET   = "2222:0000:0000:0000:0000:0000:0000:0000/64"


def config():
    return """
sp ipv6 out esp protect 5 pri 1 \\
src {0} \\
dst {1} \\
sport 0:65535 dport 0:65535

sp ipv6 in esp protect 6 pri 1 \\
src {1} \\
dst {0} \\
sport 0:65535 dport 0:65535

sa out 5 cipher_algo null auth_algo null mode transport
sa in 6 cipher_algo null auth_algo null mode transport

rt ipv6 dst {0} port 1
rt ipv6 dst {1} port 0
""".format(SRC_NET, DST_NET)


class TestTransportWithIPv6Ext(unittest.TestCase):
    # There is a bug in the IPsec Scapy implementation
    # which causes invalid packet reconstruction after
    # successful decryption. This method is a workaround.
    @staticmethod
    def decrypt(pkt, sa):
        esp = pkt[ESP]

        # decrypt dummy packet with no extensions
        d = sa.decrypt(IPv6()/esp)

        # fix 'next header' in the preceding header of the original
        # packet and remove ESP
        pkt[ESP].underlayer.nh = d[IPv6].nh
        pkt[ESP].underlayer.remove_payload()

        # combine L3 header with decrypted payload
        npkt = pkt/d[IPv6].payload

        # fix length
        npkt[IPv6].plen = d[IPv6].plen + len(pkt[IPv6].payload)

        return npkt

    def setUp(self):
        self.px = pkttest.PacketXfer()
        self.outb_sa = SecurityAssociation(ESP, spi=5)
        self.inb_sa = SecurityAssociation(ESP, spi=6)

    def test_outb_ipv6_noopt(self):
        pkt = IPv6(src=SRC_ADDR, dst=DST_ADDR)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")

        # send and check response
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_ESP)
        self.assertEqual(resp[ESP].spi, 5)

        # decrypt response, check packet after decryption
        d = TestTransportWithIPv6Ext.decrypt(resp[IPv6], self.outb_sa)
        self.assertEqual(d[IPv6].nh, socket.IPPROTO_UDP)
        self.assertEqual(d[UDP].sport, 123)
        self.assertEqual(d[UDP].dport, 456)
        self.assertEqual(bytes(d[UDP].payload), b'abc')

    def test_outb_ipv6_opt(self):
        hoptions = []
        hoptions.append(RouterAlert(value=2))
        hoptions.append(Jumbo(jumboplen=5000))
        hoptions.append(Pad1())

        doptions = []
        doptions.append(HAO(hoa="1234::4321"))

        pkt = IPv6(src=SRC_ADDR, dst=DST_ADDR)
        pkt /= IPv6ExtHdrHopByHop(options=hoptions)
        pkt /= IPv6ExtHdrRouting(addresses=["3333::3","4444::4"])
        pkt /= IPv6ExtHdrDestOpt(options=doptions)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")

        # send and check response
        resp = self.px.xfer_unprotected(pkt)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_HOPOPTS)

        # check extensions
        self.assertEqual(resp[IPv6ExtHdrHopByHop].nh, socket.IPPROTO_ROUTING)
        self.assertEqual(resp[IPv6ExtHdrRouting].nh, socket.IPPROTO_DSTOPTS)
        self.assertEqual(resp[IPv6ExtHdrDestOpt].nh, socket.IPPROTO_ESP)

        # check ESP
        self.assertEqual(resp[ESP].spi, 5)

        # decrypt response, check packet after decryption
        d = TestTransportWithIPv6Ext.decrypt(resp[IPv6], self.outb_sa)
        self.assertEqual(d[IPv6].nh, socket.IPPROTO_HOPOPTS)
        self.assertEqual(d[IPv6ExtHdrHopByHop].nh, socket.IPPROTO_ROUTING)
        self.assertEqual(d[IPv6ExtHdrRouting].nh, socket.IPPROTO_DSTOPTS)
        self.assertEqual(d[IPv6ExtHdrDestOpt].nh, socket.IPPROTO_UDP)

        # check UDP
        self.assertEqual(d[UDP].sport, 123)
        self.assertEqual(d[UDP].dport, 456)
        self.assertEqual(bytes(d[UDP].payload), b'abc')

    def test_inb_ipv6_noopt(self):
        # encrypt and send raw UDP packet
        pkt = IPv6(src=DST_ADDR, dst=SRC_ADDR)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")
        e = self.inb_sa.encrypt(pkt)

        # send and check response
        resp = self.px.xfer_protected(e)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_UDP)

        # check UDP packet
        self.assertEqual(resp[UDP].sport, 123)
        self.assertEqual(resp[UDP].dport, 456)
        self.assertEqual(bytes(resp[UDP].payload), b'abc')

    def test_inb_ipv6_opt(self):
        hoptions = []
        hoptions.append(RouterAlert(value=2))
        hoptions.append(Jumbo(jumboplen=5000))
        hoptions.append(Pad1())

        doptions = []
        doptions.append(HAO(hoa="1234::4321"))

        # prepare packet with options
        pkt = IPv6(src=DST_ADDR, dst=SRC_ADDR)
        pkt /= IPv6ExtHdrHopByHop(options=hoptions)
        pkt /= IPv6ExtHdrRouting(addresses=["3333::3","4444::4"])
        pkt /= IPv6ExtHdrDestOpt(options=doptions)
        pkt /= UDP(sport=123,dport=456)/Raw(load="abc")
        e = self.inb_sa.encrypt(pkt)

        # self encrypted packet and check response
        resp = self.px.xfer_protected(e)
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_HOPOPTS)
        self.assertEqual(resp[IPv6ExtHdrHopByHop].nh, socket.IPPROTO_ROUTING)
        self.assertEqual(resp[IPv6ExtHdrRouting].nh, socket.IPPROTO_DSTOPTS)
        self.assertEqual(resp[IPv6ExtHdrDestOpt].nh, socket.IPPROTO_UDP)

        # check UDP
        self.assertEqual(resp[UDP].sport, 123)
        self.assertEqual(resp[UDP].dport, 456)
        self.assertEqual(bytes(resp[UDP].payload), b'abc')

    def test_inb_ipv6_frag(self):
        # prepare ESP payload
        pkt = IPv6()/UDP(sport=123,dport=456)/Raw(load="abc")
        e = self.inb_sa.encrypt(pkt)

        # craft and send inbound packet
        e = IPv6(src=DST_ADDR, dst=SRC_ADDR)/IPv6ExtHdrFragment()/e[IPv6].payload
        resp = self.px.xfer_protected(e)

        # check response
        self.assertEqual(resp[IPv6].nh, socket.IPPROTO_FRAGMENT)
        self.assertEqual(resp[IPv6ExtHdrFragment].nh, socket.IPPROTO_UDP)

        # check UDP
        self.assertEqual(resp[UDP].sport, 123)
        self.assertEqual(resp[UDP].dport, 456)
        self.assertEqual(bytes(resp[UDP].payload), b'abc')


pkttest.pkttest()



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause

import fcntl
import pkg_resources
import socket
import struct
import sys
import unittest


if sys.version_info < (3, 0):
    print("Python3 is required to run this script")
    sys.exit(1)


try:
    from scapy.all import Ether
except ImportError:
    print("Scapy module is required")
    sys.exit(1)


PKTTEST_REQ = [
    "scapy>=2.4.3",
]


def assert_requirements(req):
    """
    assert requirement is met
    req can hold a string or a list of strings
    """
    try:
        pkg_resources.require(req)
    except (pkg_resources.DistributionNotFound, pkg_resources.VersionConflict) as e:
        print("Requirement assertion: " + str(e))
        sys.exit(1)


TAP_UNPROTECTED = "dtap1"
TAP_PROTECTED = "dtap0"


class Interface(object):
    ETH_P_ALL = 3
    MAX_PACKET_SIZE = 1280
    IOCTL_GET_INFO = 0x8927
    SOCKET_TIMEOUT = 0.5
    def __init__(self, ifname):
        self.name = ifname

        # create and bind socket to specified interface
        self.s = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(Interface.ETH_P_ALL))
        self.s.settimeout(Interface.SOCKET_TIMEOUT)
        self.s.bind((self.name, 0, socket.PACKET_OTHERHOST))

        # get interface MAC address
        info = fcntl.ioctl(self.s.fileno(), Interface.IOCTL_GET_INFO,  struct.pack('256s', bytes(ifname[:15], encoding='ascii')))
        self.mac = ':'.join(['%02x' % i for i in info[18:24]])

    def __del__(self):
        self.s.close()

    def send_l3packet(self, pkt, mac):
        e = Ether(src=self.mac, dst=mac)
        self.send_packet(e/pkt)

    def send_packet(self, pkt):
        self.send_bytes(bytes(pkt))

    def send_bytes(self, bytedata):
        self.s.send(bytedata)

    def recv_packet(self):
        return Ether(self.recv_bytes())

    def recv_bytes(self):
        return self.s.recv(Interface.MAX_PACKET_SIZE)

    def get_mac(self):
        return self.mac


class PacketXfer(object):
    def __init__(self, protected_iface=TAP_PROTECTED, unprotected_iface=TAP_UNPROTECTED):
        self.protected_port = Interface(protected_iface)
        self.unprotected_port = Interface(unprotected_iface)

    def send_to_protected_port(self, pkt, remote_mac=None):
        if remote_mac is None:
            remote_mac = self.unprotected_port.get_mac()
        self.protected_port.send_l3packet(pkt, remote_mac)

    def send_to_unprotected_port(self, pkt, remote_mac=None):
        if remote_mac is None:
            remote_mac = self.protected_port.get_mac()
        self.unprotected_port.send_l3packet(pkt, remote_mac)

    def xfer_unprotected(self, pkt):
        self.send_to_unprotected_port(pkt)
        return self.protected_port.recv_packet()

    def xfer_protected(self, pkt):
        self.send_to_protected_port(pkt)
        return self.unprotected_port.recv_packet()


def pkttest():
    if len(sys.argv) == 1:
        sys.exit(unittest.main(verbosity=2))
    elif len(sys.argv) == 2:
        if sys.argv[1] == "config":
            module = __import__('__main__')
            try:
                print(module.config())
            except AttributeError:
                sys.stderr.write("Cannot find \"config()\" in a test")
                sys.exit(1)
    else:
        sys.exit(1)


if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == "check_reqs":
        assert_requirements(PKTTEST_REQ)
    else:
        print("Usage: " + sys.argv[0] + " check_reqs")



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2020 Intel Corporation
#

"""
A Python program that generates the VXLAN tunnels for this example.
"""

import argparse

DESCRIPTION = 'Table Generator'

KEY = '0xaabbccdd{0:04x}'
ACTION = 'vxlan_encap'
ETHERNET_HEADER = 'ethernet_dst_addr 0xa0a1a2a3{0:04x} ' \
    'ethernet_src_addr 0xb0b1b2b3{0:04x} ' \
    'ethernet_ethertype 0x0800'
IPV4_HEADER = 'ipv4_ver_ihl 0x45 ' \
    'ipv4_diffserv 0 ' \
    'ipv4_total_len 50 ' \
    'ipv4_identification 0 ' \
    'ipv4_flags_offset 0 ' \
    'ipv4_ttl 64 ' \
    'ipv4_protocol 17 ' \
    'ipv4_hdr_checksum 0x{1:04x} ' \
    'ipv4_src_addr 0xc0c1{0:04x} ' \
    'ipv4_dst_addr 0xd0d1{0:04x}'
UDP_HEADER = 'udp_src_port 0xe0{0:02x} ' \
    'udp_dst_port 4789 ' \
    'udp_length 30 ' \
    'udp_checksum 0'
VXLAN_HEADER = 'vxlan_flags 0 ' \
    'vxlan_reserved 0 ' \
    'vxlan_vni {0:d} ' \
    'vxlan_reserved2 0'
PORT_OUT = 'port_out {0:d}'

def ipv4_header_checksum(i):
    cksum = (0x4500 + 0x0032) + (0x0000 + 0x0000) + (0x4011 + 0x0000) + (0xc0c1 + i) + (0xd0d1 + i)
    cksum = (cksum & 0xFFFF) + (cksum >> 16)
    cksum = (cksum & 0xFFFF) + (cksum >> 16)
    cksum = ~cksum & 0xFFFF
    return cksum

def table_generate(n, p):
    for i in range(0, n):
        print("match %s action %s %s %s %s %s %s" %
              (KEY.format(i), ACTION,
               ETHERNET_HEADER.format(i),
               IPV4_HEADER.format(i, ipv4_header_checksum(i)),
               UDP_HEADER.format(i % 256),
               VXLAN_HEADER.format(i),
               PORT_OUT.format(i % p)))

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description=DESCRIPTION)

    parser.add_argument(
        '-n',
        help='number of table entries (default: 65536)',
        required=False,
        default=65536)

    parser.add_argument(
        '-p',
        help='number of network ports (default: 4)',
        required=False,
        default=4)

    args = parser.parse_args()
    table_generate(int(args.n), int(args.p))



#!/usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Robin Jarry

r'''
DPDK telemetry exporter.

It uses dynamically loaded endpoint exporters which are basic python files that
must implement two functions:

    def info() -> dict[MetricName, MetricInfo]:
        """
        Mapping of metric names to their description and type.
        """

    def metrics(sock: TelemetrySocket) -> list[MetricValue]:
        """
        Request data from sock and return it as metric values. A metric value
        is a 3-tuple: (name: str, value: any, labels: dict). Each name must be
        present in info().
        """

The sock argument passed to metrics() has a single method:

    def cmd(self, uri, arg=None) -> dict | list:
        """
        Request JSON data to the telemetry socket and parse it to python
        values.
        """

See existing endpoints for examples.

The exporter supports multiple output formats:

prometheus://ADDRESS:PORT
openmetrics://ADDRESS:PORT
  Expose the enabled endpoints via a local HTTP server listening on the
  specified address and port. GET requests on that server are served with
  text/plain responses in the prometheus/openmetrics format.

  More details:
  https://prometheus.io/docs/instrumenting/exposition_formats/#text-based-format

carbon://ADDRESS:PORT
graphite://ADDRESS:PORT
  Export all enabled endpoints to the specified TCP ADDRESS:PORT in the pickle
  carbon format.

  More details:
  https://graphite.readthedocs.io/en/latest/feeding-carbon.html#the-pickle-protocol
'''

import argparse
import importlib.util
import json
import logging
import os
import pickle
import re
import socket
import struct
import sys
import time
import typing
from http import HTTPStatus, server
from urllib.parse import urlparse

LOG = logging.getLogger(__name__)
# Use local endpoints path only when running from source
LOCAL = os.path.join(os.path.dirname(__file__), "telemetry-endpoints")
DEFAULT_LOAD_PATHS = []
if os.path.isdir(LOCAL):
    DEFAULT_LOAD_PATHS.append(LOCAL)
DEFAULT_LOAD_PATHS += [
    "/usr/local/share/dpdk/telemetry-endpoints",
    "/usr/share/dpdk/telemetry-endpoints",
]
DEFAULT_OUTPUT = "openmetrics://:9876"


def main():
    logging.basicConfig(
        stream=sys.stdout,
        level=logging.INFO,
        format="%(asctime)s %(levelname)s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="FORMAT://PARAMETERS",
        default=urlparse(DEFAULT_OUTPUT),
        type=urlparse,
        help=f"""
        Output format (default: "{DEFAULT_OUTPUT}"). Depending on the format,
        URL elements have different meanings. By default, the exporter starts a
        local HTTP server on port 9876 that serves requests in the
        prometheus/openmetrics plain text format.
        """,
    )
    parser.add_argument(
        "-p",
        "--load-path",
        dest="load_paths",
        type=lambda v: v.split(os.pathsep),
        default=DEFAULT_LOAD_PATHS,
        help=f"""
        The list of paths from which to disvover endpoints.
        (default: "{os.pathsep.join(DEFAULT_LOAD_PATHS)}").
        """,
    )
    parser.add_argument(
        "-e",
        "--endpoint",
        dest="endpoints",
        metavar="ENDPOINT",
        action="append",
        help="""
        Telemetry endpoint to export (by default, all discovered endpoints are
        enabled). This option can be specified more than once.
        """,
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help="""
        Only list detected endpoints and exit.
        """,
    )
    parser.add_argument(
        "-s",
        "--socket-path",
        default="/run/dpdk/rte/dpdk_telemetry.v2",
        help="""
        The DPDK telemetry socket path (default: "%(default)s").
        """,
    )
    args = parser.parse_args()
    output = OUTPUT_FORMATS.get(args.output.scheme)
    if output is None:
        parser.error(f"unsupported output format: {args.output.scheme}://")

    try:
        endpoints = load_endpoints(args.load_paths, args.endpoints)
        if args.list:
            return
    except Exception as e:
        parser.error(str(e))

    output(args, endpoints)


class TelemetrySocket:
    """
    Abstraction of the DPDK telemetry socket.
    """

    def __init__(self, path: str):
        self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
        self.sock.connect(path)
        data = json.loads(self.sock.recv(1024).decode())
        self.max_output_len = data["max_output_len"]

    def cmd(
        self, uri: str, arg: typing.Any = None
    ) -> typing.Optional[typing.Union[dict, list]]:
        """
        Request JSON data to the telemetry socket and parse it to python
        values.
        """
        if arg is not None:
            u = f"{uri},{arg}"
        else:
            u = uri
        self.sock.send(u.encode("utf-8"))
        data = self.sock.recv(self.max_output_len)
        return json.loads(data.decode("utf-8"))[uri]

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        self.sock.close()


MetricDescription = str
MetricType = str
MetricName = str
MetricLabels = typing.Dict[str, typing.Any]
MetricInfo = typing.Tuple[MetricDescription, MetricType]
MetricValue = typing.Tuple[MetricName, typing.Any, MetricLabels]


class TelemetryEndpoint:
    """
    Placeholder class only used for typing annotations.
    """

    @staticmethod
    def info() -> typing.Dict[MetricName, MetricInfo]:
        """
        Mapping of metric names to their description and type.
        """
        raise NotImplementedError()

    @staticmethod
    def metrics(sock: TelemetrySocket) -> typing.List[MetricValue]:
        """
        Request data from sock and return it as metric values. Each metric
        name must be present in info().
        """
        raise NotImplementedError()


def load_endpoints(
    paths: typing.List[str], names: typing.List[str]
) -> typing.List[TelemetryEndpoint]:
    """
    Load selected telemetry endpoints from the specified paths.
    """

    endpoints = {}
    dwb = sys.dont_write_bytecode
    sys.dont_write_bytecode = True  # never generate .pyc files for endpoints

    for p in paths:
        if not os.path.isdir(p):
            continue
        for fname in os.listdir(p):
            f = os.path.join(p, fname)
            if os.path.isdir(f):
                continue
            try:
                name, _ = os.path.splitext(fname)
                if names is not None and name not in names:
                    # not selected by user
                    continue
                if name in endpoints:
                    # endpoint with same name already loaded
                    continue
                spec = importlib.util.spec_from_file_location(name, f)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
                endpoints[name] = module
            except Exception:
                LOG.exception("parsing endpoint: %s", f)

    if not endpoints:
        raise Exception("no telemetry endpoints detected/selected")

    sys.dont_write_bytecode = dwb

    modules = []
    info = {}
    for name, module in sorted(endpoints.items()):
        LOG.info("using endpoint: %s (from %s)", name, module.__file__)
        try:
            for metric, (description, type_) in module.info().items():
                info[(name, metric)] = (description, type_)
            modules.append(module)
        except Exception:
            LOG.exception("getting endpoint info: %s", name)
    return modules


def serve_openmetrics(
    args: argparse.Namespace, endpoints: typing.List[TelemetryEndpoint]
):
    """
    Start an HTTP server and serve requests in the openmetrics/prometheus
    format.
    """
    listen = (args.output.hostname or "", int(args.output.port or 80))
    with server.HTTPServer(listen, OpenmetricsHandler) as httpd:
        httpd.dpdk_socket_path = args.socket_path
        httpd.telemetry_endpoints = endpoints
        LOG.info("listening on port %s", httpd.server_port)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            LOG.info("shutting down")


class OpenmetricsHandler(server.BaseHTTPRequestHandler):
    """
    Basic HTTP handler that returns prometheus/openmetrics formatted responses.
    """

    CONTENT_TYPE = "text/plain; version=0.0.4; charset=utf-8"

    def escape(self, value: typing.Any) -> str:
        """
        Escape a metric label value.
        """
        value = str(value)
        value = value.replace('"', '\\"')
        value = value.replace("\\", "\\\\")
        return value.replace("\n", "\\n")

    def do_GET(self):
        """
        Called upon GET requests.
        """
        try:
            lines = []
            metrics_names = set()
            with TelemetrySocket(self.server.dpdk_socket_path) as sock:
                for e in self.server.telemetry_endpoints:
                    info = e.info()
                    metrics_lines = []
                    try:
                        metrics = e.metrics(sock)
                    except Exception:
                        LOG.exception("%s: metrics collection failed", e.__name__)
                        continue
                    for name, value, labels in metrics:
                        fullname = re.sub(r"\W", "_", f"dpdk_{e.__name__}_{name}")
                        labels = ", ".join(
                            f'{k}="{self.escape(v)}"' for k, v in labels.items()
                        )
                        if labels:
                            labels = f"{{{labels}}}"
                        metrics_lines.append(f"{fullname}{labels} {value}")
                        if fullname not in metrics_names:
                            metrics_names.add(fullname)
                            desc, metric_type = info[name]
                            lines += [
                                f"# HELP {fullname} {desc}",
                                f"# TYPE {fullname} {metric_type}",
                            ]
                    lines += metrics_lines
            if not lines:
                self.send_error(HTTPStatus.INTERNAL_SERVER_ERROR)
                LOG.error(
                    "%s %s: no metrics collected",
                    self.address_string(),
                    self.requestline,
                )
            body = "\n".join(lines).encode("utf-8") + b"\n"
            self.send_response(HTTPStatus.OK)
            self.send_header("Content-Type", self.CONTENT_TYPE)
            self.send_header("Content-Length", str(len(body)))
            self.end_headers()
            self.wfile.write(body)
            LOG.info("%s %s", self.address_string(), self.requestline)

        except (FileNotFoundError, ConnectionRefusedError):
            self.send_error(HTTPStatus.SERVICE_UNAVAILABLE)
            LOG.exception(
                "%s %s: telemetry socket not available",
                self.address_string(),
                self.requestline,
            )
        except Exception:
            self.send_error(HTTPStatus.INTERNAL_SERVER_ERROR)
            LOG.exception("%s %s", self.address_string(), self.requestline)

    def log_message(self, fmt, *args):
        pass  # disable built-in logger


def export_carbon(args: argparse.Namespace, endpoints: typing.List[TelemetryEndpoint]):
    """
    Collect all metrics and export them to a carbon server in the pickle format.
    """
    addr = (args.output.hostname or "", int(args.output.port or 80))
    with TelemetrySocket(args.socket_path) as dpdk:
        with socket.socket() as carbon:
            carbon.connect(addr)
            all_metrics = []
            for e in endpoints:
                try:
                    metrics = e.metrics(dpdk)
                except Exception:
                    LOG.exception("%s: metrics collection failed", e.__name__)
                    continue
                for name, value, labels in metrics:
                    fullname = re.sub(r"\W", ".", f"dpdk.{e.__name__}.{name}")
                    for key, val in labels.items():
                        val = str(val).replace(";", "")
                        fullname += f";{key}={val}"
                    all_metrics.append((fullname, (time.time(), value)))
            if not all_metrics:
                raise Exception("no metrics collected")
            payload = pickle.dumps(all_metrics, protocol=2)
            header = struct.pack("!L", len(payload))
            buf = header + payload
            carbon.sendall(buf)


OUTPUT_FORMATS = {
    "openmetrics": serve_openmetrics,
    "prometheus": serve_openmetrics,
    "carbon": export_carbon,
    "graphite": export_carbon,
}


if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# SPDX-License-Identifier: BSD-3-Clause
# Copyright(c) 2020 Intel Corporation

"""
Script to be used with V2 Telemetry.
Allows the user input commands and read the Telemetry response.
"""

import socket
import os
import sys
import glob
import json
import errno
import readline
import argparse

# global vars
TELEMETRY_VERSION = "v2"
SOCKET_NAME = 'dpdk_telemetry.{}'.format(TELEMETRY_VERSION)
DEFAULT_PREFIX = 'rte'
CMDS = []


def read_socket(sock, buf_len, echo=True, pretty=False):
    """ Read data from socket and return it in JSON format """
    reply = sock.recv(buf_len).decode()
    try:
        ret = json.loads(reply)
    except json.JSONDecodeError:
        print("Error in reply: ", reply)
        sock.close()
        raise
    if echo:
        indent = 2 if pretty else None
        print(json.dumps(ret, indent=indent))
    return ret


def get_app_name(pid):
    """ return the app name for a given PID, for printing """
    proc_cmdline = os.path.join('/proc', str(pid), 'cmdline')
    try:
        with open(proc_cmdline) as f:
            argv0 = f.read(1024).split('\0')[0]
            return os.path.basename(argv0)
    except IOError as e:
        # ignore file not found errors
        if e.errno != errno.ENOENT:
            raise
    return None


def find_sockets(path):
    """ Find any possible sockets to connect to and return them """
    return glob.glob(os.path.join(path, SOCKET_NAME + '*'))


def print_socket_options(prefix, paths):
    """ Given a set of socket paths, give the commands needed to connect """
    cmd = sys.argv[0]
    if prefix != DEFAULT_PREFIX:
        cmd += " -f " + prefix
    for s in sorted(paths):
        sock_name = os.path.basename(s)
        if sock_name.endswith(TELEMETRY_VERSION):
            print("- {}  # Connect with '{}'".format(os.path.basename(s),
                                                     cmd))
        else:
            print("- {}  # Connect with '{} -i {}'".format(os.path.basename(s),
                                                           cmd,
                                                           s.split(':')[-1]))


def get_dpdk_runtime_dir(fp):
    """ Using the same logic as in DPDK's EAL, get the DPDK runtime directory
    based on the file-prefix and user """
    run_dir = os.environ.get('RUNTIME_DIRECTORY')
    if not run_dir:
        if (os.getuid() == 0):
            run_dir = '/var/run'
        else:
            run_dir = os.environ.get('XDG_RUNTIME_DIR', '/tmp')
    return os.path.join(run_dir, 'dpdk', fp)


def list_fp():
    """ List all available file-prefixes to user """
    path = get_dpdk_runtime_dir('')
    sockets = glob.glob(os.path.join(path, "*", SOCKET_NAME + "*"))
    prefixes = []
    if not sockets:
        print("No DPDK apps with telemetry enabled available")
    else:
        print("Valid file-prefixes:\n")
    for s in sockets:
        prefixes.append(os.path.relpath(os.path.dirname(s), start=path))
    for p in sorted(set(prefixes)):
        print(p)
        print_socket_options(p, glob.glob(os.path.join(path, p,
                                                       SOCKET_NAME + "*")))


def handle_socket(args, path):
    """ Connect to socket and handle user input """
    prompt = ''  # this evaluates to false in conditions
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_SEQPACKET)
    global CMDS

    if os.isatty(sys.stdin.fileno()):
        prompt = '--> '
        print("Connecting to " + path)
    try:
        sock.connect(path)
    except OSError:
        print("Error connecting to " + path)
        sock.close()
        # if socket exists but is bad, or if non-interactive just return
        if os.path.exists(path) or not prompt:
            return
        # if user didn't give a valid socket path, but there are
        # some sockets, help the user out by printing how to connect
        socks = find_sockets(os.path.dirname(path))
        if socks:
            print("\nOther DPDK telemetry sockets found:")
            print_socket_options(args.file_prefix, socks)
        else:
            list_fp()
        return
    json_reply = read_socket(sock, 1024, prompt, prompt)
    output_buf_len = json_reply["max_output_len"]
    app_name = get_app_name(json_reply["pid"])
    if app_name and prompt:
        print('Connected to application: "%s"' % app_name)

    # get list of commands for readline completion
    sock.send("/".encode())
    CMDS = read_socket(sock, output_buf_len, False)["/"]

    # interactive prompt
    try:
        text = input(prompt).strip()
        while text != "quit":
            if text.startswith('/'):
                sock.send(text.encode())
                read_socket(sock, output_buf_len, pretty=prompt)
            text = input(prompt).strip()
    except EOFError:
        pass
    finally:
        sock.close()


def readline_complete(text, state):
    """ Find any matching commands from the list based on user input """
    all_cmds = ['quit'] + CMDS
    if text:
        matches = [c for c in all_cmds if c.startswith(text)]
    else:
        matches = all_cmds
    return matches[state]


readline.parse_and_bind('tab: complete')
readline.set_completer(readline_complete)
readline.set_completer_delims(readline.get_completer_delims().replace('/', ''))

parser = argparse.ArgumentParser()
parser.add_argument('-f', '--file-prefix', default=DEFAULT_PREFIX,
                    help='Provide file-prefix for DPDK runtime directory')
parser.add_argument('-i', '--instance', default='0', type=int,
                    help='Provide instance number for DPDK application')
parser.add_argument('-l', '--list', action="store_true", default=False,
                    help='List all possible file-prefixes and exit')
args = parser.parse_args()
if args.list:
    list_fp()
    sys.exit(0)
sock_path = os.path.join(get_dpdk_runtime_dir(args.file_prefix), SOCKET_NAME)
if args.instance > 0:
    sock_path += ":{}".format(args.instance)
handle_socket(args, sock_path)
